"""Module containing the logic for console command line usage"""

import sys
import re
from enum import IntFlag

from dgspoc.utils import Printer
from dgspoc.utils import Misc

from dgspoc.constant import ECODE

from dgspoc import example
from dgspoc.example import get_number_of_example

tool = 'dgs'


class FLAG(IntFlag):
    AUTHOR = pow(2, 0)
    EMAIL = pow(2, 1)
    COMPANY = pow(2, 2)
    SAVE_TO = pow(2, 3)
    TEMPLATE_ID = pow(2, 4)
    ADAPTOR = pow(2, 5)
    ACTION = pow(2, 6)
    REPLACED = pow(2, 7)
    ALL = pow(2, 8)
    DEPENDENCY = pow(2, 9)
    TEMPLATE_STORAGE = pow(2, 10)
    HELP = pow(2, 11)

    BUILD_TEMPLATE = AUTHOR | EMAIL | COMPANY | SAVE_TO | TEMPLATE_ID | REPLACED | HELP
    SEARCH_TEMPLATE = HELP
    INFO_USAGE = ALL | DEPENDENCY | TEMPLATE_STORAGE | HELP
    TEST_USAGE = ADAPTOR | ACTION | HELP


class UData:
    def __init__(self, *args, is_header=False):
        self.args = args
        lst = []
        if self.args:
            for arg in self.args:
                if Misc.is_list(arg):
                    lst.extend([str(item) for item in arg])
                else:
                    lst.append(str(arg))
            self.data = '\n'.join(lst)
            if is_header:
                self.data = '{0}\n{1}\n{0}'.format('+' * 80, self.data)
        else:
            self.data = ''

        if not self.data.strip():
            self.data = self.data.strip()

        self.data_len = len(self.data)

        self._count = len(lst)

    def __len__(self):
        return self.data_len

    def __repr__(self):
        return self.data

    def __str__(self):
        return self.data

    @property
    def count(self):
        return self._count


class UsageData(UData):
    def __init__(self, header_data, body_data):
        super().__init__(header_data, '{}\n'.format(body_data), is_header=False)
        self._count = body_data.count


class UHeaderData(UData):
    def __init__(self, *args):
        if len(args) > 1:
            item0 = args[0]
            lst = [item0, '-' * len(str(item0)), *args[1:]]
        else:
            lst = args
        super().__init__(*lst, is_header=True)


class UBodyData(UData):
    def __init__(self, *args):
        super().__init__(*args, is_header=False)


def get_usage_header(name, flags=0):
    name = str(name).lower()
    lst = ['{} {} usage'.format(tool, name.replace('_', ' '))]
    args = [
        "  --author AUTHOR         author's name",
        "  --email EMAIL           author's email",
        "  --company COMPANY       author's company",
        '  --save-to FILENAME      saving to file',
        '  --template-id TMPLID    template ID',
        '  --adaptor ADAPTOR       connector adaptor',
        '  --action ACTION         execution action which uses to test template or verification',
        '  --replaced              overwrite template ID/file',
        '  --all                   showing all information',
        '  --dependency            showing package dependencies',
        '  --template-storage      showing template storage information',
        '  -h, --help              show this help message and exit',
    ]
    if flags:
        bits = list(map(int, list(bin(int(flags))[2:][::-1])))
        lst.append('optional arguments:')
        lst.append('-------------------')
        for index, bit in enumerate(bits):
            bit and lst.append(args[index])

    header_usage = UHeaderData(lst)
    return header_usage


def get_usage(name, flags=0):
    count = get_number_of_example(name)
    name = str(name).lower()
    header_usage = get_usage_header(name, flags=flags)

    lst = ['{} {} operands [options]'.format(tool, name.replace('_', ' '))]
    if count > 0:
        lst1 = list(map(str, range(1, count + 1)))
        s = lst1[0] if len(lst1) == 1 else '{%s}' % (','.join(lst1))
        lst.append('%s %s example %s' % (tool, name.replace('_', ' '), s))

    body_usage = UBodyData(*lst)

    usage = UsageData(header_usage, body_usage)
    return usage


def get_example_usage(name):
    count = get_number_of_example(name)
    name = str(name).lower()
    fmt = '{} {} example {}'

    example_usage = UsageData(
        UHeaderData('{} {} example syntax:'.format(tool, name.replace('_', ' '))),
        UBodyData(*[fmt.format(tool, name, i + 1) for i in range(count)])
    )
    return example_usage


class BuildTemplateUsage:
    usage = get_usage('build_template', flags=FLAG.BUILD_TEMPLATE)
    other_usage = get_usage('build_template', flags=FLAG.BUILD_TEMPLATE)
    example_usage = get_example_usage('build_template')


class BuildUsage:
    usage = '\n'.join([
        Printer.get('build command has two features: template or script'),
        str(BuildTemplateUsage.usage),
        '',
        # BuildScriptUsage.usage,
    ])


class InfoUsage:
    usage = get_usage('info', flags=FLAG.INFO_USAGE)
    other_usage = get_usage('info', flags=FLAG.INFO_USAGE)
    example_usage = get_example_usage('info')


class SearchTemplateUsage:
    usage = get_usage('search_template', flags=FLAG.SEARCH_TEMPLATE)
    other_usage = get_usage('search_template', flags=FLAG.SEARCH_TEMPLATE)
    example_usage = get_example_usage('search_template')


class TestUsage:
    usage = get_usage('test', flags=FLAG.TEST_USAGE)
    other_usage = get_usage('test', flags=FLAG.TEST_USAGE)
    example_usage = get_example_usage('test')


class Usage:
    info = InfoUsage
    build = BuildUsage
    build_template = BuildTemplateUsage
    search_template = SearchTemplateUsage
    test = TestUsage


def validate_usage(name, operands):
    result = ''.join(operands) if Misc.is_list(operands) else str(operands)
    if result.strip().lower() == 'usage':
        show_usage(name, exit_code=ECODE.SUCCESS)


def show_usage(name, *args, exit_code=None):
    obj = getattr(Usage, name, None)
    if getattr(obj, 'usage', None):
        attr = '_'.join(list(args) + ['usage'])
        print(getattr(obj, attr))
        Misc.is_integer(exit_code) and sys.exit(exit_code)
    else:
        fmt = '*** ErrorUsage: "{}" has not defined or unavailable.'
        print(fmt.format(name))
        sys.exit(ECODE.BAD)


def validate_example_usage(name, operands):
    max_count = get_number_of_example(name)
    pattern = r'example *(?P<index>[0-9]+)$'
    txt = ' '.join(operands).strip().lower()
    m = re.match(pattern, txt)
    if m:
        index = m.group('index')
        if 1 <= int(index) <= max_count:
            cls_name = '{}Example'.format(name.title().replace('_', ''))
            cls = getattr(example, cls_name)
            result = cls.get(str(index))
            print('\n\n{}\n'.format(result))
            sys.exit(ECODE.SUCCESS)
        else:
            show_usage(name, 'example', exit_code=ECODE.BAD)
    else:
        if re.match('example', txt):
            show_usage(name, 'example', exit_code=ECODE.BAD)


def get_global_usage():
    lst = [
        UHeaderData('{} other usages'.format(tool)),
        UBodyData(
            '{} version'.format(tool),
        ),
        '',
        InfoUsage.usage,
        BuildTemplateUsage.usage,
        SearchTemplateUsage.usage,
        TestUsage.usage,
    ]

    return '\n'.join(str(item) for item in lst)



"""Module containing the logic for describe-get-system proof of conception entry-points."""

import sys
import re
import argparse

from dgspoc import version
from dgspoc.config import Data

from dgspoc.utils import Printer
from dgspoc.utils import Text

from dgspoc.constant import ECODE

from dgspoc.usage import get_global_usage
from dgspoc.usage import validate_usage
# from dgspoc.usage import show_usage
from dgspoc.usage import validate_example_usage

from dgspoc.operation import do_clear_template
from dgspoc.operation import do_build_template
from dgspoc.operation import do_search_template
from dgspoc.operation import do_testing


class ArgumentParser(argparse.ArgumentParser):

    def parse_args(self, *args, **kwargs):
        try:
            options = super().parse_args(*args, **kwargs)
        except BaseException as ex:    # noqa
            if isinstance(ex, SystemExit):
                if ex.code == ECODE.SUCCESS:
                    sys.exit(ECODE.SUCCESS)
                else:
                    self.print_help()
                    sys.exit(ECODE.BAD)
            else:
                Printer.print_message('\n{}\n', Text(ex))
                self.print_help()
                sys.exit(ECODE.BAD)

        if options.help:
            if not options.command:
                self.print_help()
                sys.exit(ECODE.SUCCESS)
            else:
                if options.command in Cli.commands:
                    command = options.command
                    feature = options.operands[0].lower() if options.operands else ''
                    name = '{}_{}'.format(command, feature) if feature else command
                    validate_usage(name, ['usage'])
                else:
                    self.print_help()
                    sys.exit(ECODE.BAD)
        return options


def show_info(options):
    command, operands = options.command, options.operands
    if command == 'info':
        name = command
        validate_usage(command, operands)
        validate_example_usage(name, operands)

        op_txt = ' '.join(operands).lower()

        lst = []
        default_lst = [
            'Describe-Get-System Proof of Concept',
            Data.get_app_info()
        ]

        is_showed_all = options.all or re.search('all', op_txt)
        is_showed_dependency = options.dependency or re.search('depend', op_txt)
        is_showed_storage = options.template_storage or re.search('template|storage', op_txt)

        if is_showed_all:
            lst.extend(default_lst)

        if is_showed_all or is_showed_dependency:
            lst and lst.append('--------------------')
            lst.append('Packages:')
            values = Data.get_dependency().values()
            for pkg in sorted(values, key=lambda item: item.get('package')):
                lst.append('  + Package: {0[package]}'.format(pkg))
                lst.append('             {0[url]}'.format(pkg))

        if is_showed_all or is_showed_storage:
            lst and lst.append('--------------------', )
            lst.append(Data.get_template_storage_info())

        Printer.print(lst) if lst else Printer.print(default_lst)
        sys.exit(ECODE.SUCCESS)


def show_version(options):
    if options.command == 'version':
        print('{} v{}'.format(Cli.prog, version))
        sys.exit(ECODE.SUCCESS)


def show_global_usage(options):
    if options.command == 'usage':
        print(get_global_usage())
        sys.exit(ECODE.SUCCESS)


class Cli:
    """describe-get-system proof of concept console CLI application."""
    prog = 'dgs'
    prog_fn = 'describe-get-system'
    commands = ['build', 'info', 'run', 'search', 'test', 'version', 'usage']

    def __init__(self):
        # parser = argparse.ArgumentParser(
        parser = ArgumentParser(
            prog=self.prog,
            usage='%(prog)s [options] command operands',
            description='{} Proof of Concept'.format(self.prog_fn.title()),
            add_help=False
        )

        parser.add_argument(
            '-h', '--help', action='store_true',
            help='show this help message and exit'
        )

        parser.add_argument(
            '-v', '--version', action='version',
            version='%(prog)s v{}'.format(version)
        )

        parser.add_argument(
            '--author', type=str, default='',
            help="author's name"
        ),

        parser.add_argument(
            '--email', type=str, default='',
            help="author's email"
        ),

        parser.add_argument(
            '--company', type=str, default='',
            help="author's company"
        ),

        parser.add_argument(
            '--save-to', type=str, dest='filename', default='',
            help="saving to file"
        ),

        parser.add_argument(
            '--template-id', type=str, dest='tmplid', default='',
            help="template ID"
        ),

        parser.add_argument(
            '--clear', type=str, dest='template_id', default='',
            help="clear template from template-storage"
        ),

        parser.add_argument(
            '--adaptor', type=str, default='',
            help="connector adaptor"
        ),

        parser.add_argument(
            '--action', type=str, default='',
            help="execution action which uses to test template or verification"
        ),

        parser.add_argument(
            '--replaced', action='store_true',
            help='overwrite template ID/file'
        )

        parser.add_argument(
            '--all', action='store_true',
            help='showing all information'
        )

        parser.add_argument(
            '--dependency', action='store_true',
            help='showing package dependency'
        )

        parser.add_argument(
            '--template-storage', action='store_true',
            help='showing template storage information'
        )

        parser.add_argument(
            'command', nargs='?', type=str, default='',
            help='command must be either build, '
                 'info, run, search, test, version, or usage'
        )
        parser.add_argument(
            'operands', nargs='*', type=str,
            help='operands can be template, unittest, '
                 'pytest, robotframework, script, or data such command-line, '
                 'config-lines, or filename'
        )

        self.kwargs = dict()
        self.parser = parser
        self.options = self.parser.parse_args()

    def validate_command(self):
        """Validate argparse `options.command`.

        Returns
        -------
        bool: show ``self.parser.print_help()`` and call ``sys.exit(ECODE.BAD)`` if
        command is neither build, info, run, search,
        test, version, nor usage otherwise, return True
        """
        self.options.command = self.options.command.lower()

        if self.options.command:
            if self.options.command in self.commands:
                return True
            else:
                self.parser.print_help()
                sys.exit(ECODE.BAD)
        return True

    def run(self):
        """Take CLI arguments, parse it, and process."""
        self.validate_command()

        show_version(self.options)
        show_global_usage(self.options)
        show_info(self.options)

        # operation
        do_clear_template(self.options)
        do_build_template(self.options)
        do_search_template(self.options)

        do_testing(self.options)


def execute():
    """Execute template console CLI."""
    app = Cli()
    app.run()



"""Module containing the logic for utilities."""

import re

import subprocess
from copy import deepcopy

from pathlib import Path
from pathlib import PurePath
from pathlib import WindowsPath
from datetime import datetime

from textwrap import wrap

from argparse import ArgumentParser

import yaml

import typing

from io import StringIO

from textfsm import TextFSM

from dgspoc.exceptions import UtilsParsedTemplateError

from dgspoc.constant import ECODE


class Text(str):
    def __new__(cls, *args, **kwargs):
        if not args and not kwargs:
            return str.__new__(cls, '')
        encoding = kwargs.get('encoding', 'utf-8')
        errors = kwargs.get('errors', 'strict')
        obj = kwargs.get('object', '')
        if args:
            if len(args) == 1:
                obj = args[0]
                if isinstance(obj, bytes):
                    return str.__new__(cls, obj, encoding=encoding, errors=errors)
                elif isinstance(obj, BaseException):
                    return str.__new__(cls, '{}: {}'.format(type(obj).__name__, obj))
                else:
                    return str.__new__(cls, obj)
            else:
                return str.__new__(cls, *args, **kwargs)
        else:
            if isinstance(obj, bytes):
                return str.__new__(cls, obj, encoding=encoding, errors=errors)
            elif isinstance(obj, BaseException):
                return str.__new__(cls, '{}: {}'.format(type(obj).__name__, obj))
            else:
                return str.__new__(cls, obj)


class Printer:
    """A printer class.

    Methods
    Printer.get(data, header='', footer='', failure_msg='', width=80, width_limit=20) -> str
    Printer.print(data, header='', footer='', failure_msg='', width=80, width_limit=20, print_func=None) -> None
    """
    @classmethod
    def get(cls, data, header='', footer='',
            width=80, width_limit=20, failure_msg=''):
        """Decorate data by organizing header, data, footer, and failure_msg

        Parameters
        ----------
        data (str, list): a text or a list of text.
        header (str): a header text.  Default is empty.
        footer (str): a footer text.  Default is empty.
        width (int): width of displayed text.  Default is 80.
        width_limit (int): minimum width of displayed text.  Default is 20.
        failure_msg (str): a failure message.  Default is empty.
        """
        headers = str(header).splitlines()
        footers = str(footer).splitlines()
        data = data if Misc.is_mutable_sequence(data) else [data]
        lst = []
        result = []

        right_bound = width - 4

        for item in data:
            if width >= width_limit:
                for line in str(item).splitlines():
                    lst.extend(wrap(line, width=right_bound))
            else:
                lst.extend(line.rstrip() for line in str(item).splitlines())
        length = max(len(str(i)) for i in lst + headers + footers)

        if width >= width_limit:
            length = right_bound if right_bound > length else length

        result.append('+-{}-+'.format('-' * length))
        if header:
            for item in headers:
                result.append('| {} |'.format(item.ljust(length)))
            result.append('+-{}-+'.format('-' * length))

        for item in lst:
            result.append('| {} |'.format(item.ljust(length)))
        result.append('+-{}-+'.format('-' * length))

        if footer:
            for item in footers:
                result.append('| {} |'.format(item.ljust(length)))
            result.append('+-{}-+'.format('-' * length))

        if failure_msg:
            result.append(failure_msg)

        txt = '\n'.join(result)
        return txt

    @classmethod
    def print(cls, data, header='', footer='',
              width=80, width_limit=20, failure_msg='', print_func=None):
        """Decorate data by organizing header, data, footer, and failure_msg

        Parameters
        ----------
        data (str, list): a text or a list of text.
        header (str): a header text.  Default is empty.
        footer (str): a footer text.  Default is empty.
        width (int): width of displayed text.  Default is 80.
        width_limit (int): minimum width of displayed text.  Default is 20.
        failure_msg (str): a failure message.  Default is empty.
        print_func (function): a print function.  Default is None.
        """

        txt = Printer.get(data, header=header, footer=footer,
                          failure_msg=failure_msg, width=width,
                          width_limit=width_limit)

        print_func = print_func if callable(print_func) else print
        print_func(txt)

    @classmethod
    def get_message(cls, fmt, *args, style='format', prefix=''):
        """Get a message

        Parameters
        ----------
        fmt (str): string format.
        args (tuple): list of parameters for string interpolation.
        style (str): either format or %.
        prefix (str): a prefix.

        Returns
        -------
        str: a message.
        """

        if args:
            message = fmt.format(*args) if style == 'format' else fmt % args
        else:
            message = fmt

        message = '{} {}'.format(prefix, message) if prefix else message
        return message

    @classmethod
    def print_message(cls, fmt, *args, style='format', prefix='', print_func=None):
        """Print a message

        Parameters
        ----------
        fmt (str): string format.
        args (tuple): list of parameters for string interpolation.
        style (str): either format or %.
        prefix (str): a prefix.
        print_func (function): a print function.
        """
        message = cls.get_message(fmt, *args, style=style, prefix=prefix)
        print_func = print_func if callable(print_func) else print
        print_func(message)


class File:
    message = ''

    @classmethod
    def is_exist(cls, filename):
        """Check file existence

        Parameters
        ----------
        filename (str): a file name

        Returns
        -------
        bool: True if existed, otherwise False
        """
        try:
            file_obj = Path(filename)
            return file_obj.exists()
        except Exception as ex:
            cls.message = Text(ex)
            return False

    @classmethod
    def create(cls, filename, showed=True):
        """Check file existence

        Parameters
        ----------
        filename (str): a file name
        showed (bool): showing the message of creating file

        Returns
        -------
        bool: True if created, otherwise False
        """
        filename = cls.get_path(str(filename).strip())
        if cls.is_exist(filename):
            cls.message = 'File is already existed.'
            return True

        try:
            file_obj = Path(filename)
            if not file_obj.parent.exists():
                file_obj.parent.mkdir(parents=True, exist_ok=True)
            file_obj.touch()
            fmt = '{:%Y-%m-%d %H:%M:%S.%f} - {} file is created.'
            showed and print(fmt.format(datetime.now(), filename))
            cls.message = '{} file is created.'.format(filename)
            return True
        except Exception as ex:
            cls.message = Text(ex)
            return False

    @classmethod
    def get_path(cls, *args, is_home=False):
        """Create a file path

        Parameters
        ----------
        args (tuple): a list of file items
        is_home (bool): True will include Home directory.  Default is False.

        Returns
        -------
        str: a file path.
        """
        lst = [Path.home()] if is_home else []
        lst.extend(list(args))
        file_path = str(Path(PurePath(*lst)).expanduser().absolute())
        return file_path

    @classmethod
    def get_dir(cls, file_path):
        """get directory from existing file path

        Parameters
        ----------
        file_path (string): file path

        Returns
        -------
        str: directory
        """
        file_obj = Path(file_path).expanduser().absolute()
        if file_obj.is_dir():
            return str(file_obj)
        elif file_obj.is_file():
            return str(file_obj.parent)
        else:
            fmt = 'FileNotFoundError: No such file or directory "{}"'
            cls.message = fmt.format(file_path)
            return ''

    @classmethod
    def get_content(cls, file_path):
        """get content of file

        Parameters
        ----------
        file_path (string): file path

        Returns
        -------
        str: content of file
        """
        filename = cls.get_path(file_path)
        try:
            with open(filename) as stream:
                content = stream.read()
                return content
        except Exception as ex:
            cls.message = Text(ex)
            return ''

    @classmethod
    def get_result_from_yaml_file(cls, file_path, base_dir='',
                                  is_stripped=True,
                                  dot_datatype=False,
                                  default=dict(),  # noqa
                                  var_substitution=False,
                                  root_var_name='self'
                                  ):
        """get result of YAML file

        Parameters
        ----------
        file_path (string): file path
        base_dir (str): a based directory
        is_stripped (bool): removing leading or trailing space.  Default is True.
        dot_datatype (bool): convert a return_result to DotObject if
                return_result is dictionary.  Default is False.
        default (object): a default result file is not found.  Default is empty dict.
        var_substitution (bool): internal variable substitution.  Default is False.
        root_var_name (str): root variable of data structure for
                variable substitution.  Default is self.

        Returns
        -------
        object: YAML result
        """

        yaml_result = default

        try:
            if base_dir:
                filename = cls.get_path(cls.get_dir(base_dir), file_path)
            else:
                filename = cls.get_path(file_path)

            with open(filename) as stream:
                content = stream.read()
                if is_stripped:
                    content = content.strip()

                if content:
                    yaml_result = yaml.safe_load(content)
                    cls.message = 'loaded {}'.format(filename)
                else:
                    cls.message = '"{}" file is empty.'.format(filename)

        except Exception as ex:
            cls.message = Text(ex)

        if var_substitution:
            yaml_result = Misc.substitute_variable(yaml_result,
                                                   root_var_name=root_var_name)

        if Misc.is_dict(yaml_result) and dot_datatype:
            dot_result = DotObject(yaml_result)
            return dot_result
        else:
            return yaml_result

    @classmethod
    def save(cls, filename, data):
        """save data to file

        Parameters
        ----------
        filename (str): filename
        data (str): data.

        Returns
        -------
        bool: True if successfully saved, otherwise, False
        """
        try:
            if Misc.is_list(data):
                content = '\n'.join(str(item) for item in data)
            else:
                content = str(data)

            filename = cls.get_path(filename)
            if not cls.create(filename):
                return False

            file_obj = Path(filename)
            file_obj.touch()
            file_obj.write_text(content)
            cls.message = 'Successfully saved data to "{}" file'.format(filename)
            return True
        except Exception as ex:
            cls.message = Text(ex)
            return False

    @classmethod
    def delete(cls, filename):
        """Delete file

        Parameters
        ----------
        filename (str): filename

        Returns
        -------
        bool: True if successfully deleted, otherwise, False
        """
        try:
            filepath = File.get_path(filename)
            file_obj = Path(filepath)
            if file_obj.is_dir():
                file_obj.rmdir()
            else:
                file_obj.unlink()
            cls.message = 'Successfully deleted "{}" file'.format(filename)
            return True
        except Exception as ex:
            cls.message = Text(ex)
            return False

    @classmethod
    def change_home_dir_to_generic(cls, filename):
        """change HOME DIRECTORY in filename to generic name
        ++++++++++++++++++++++++++++++++++++++++++++++
        Note: this function only uses for displaying.
        ++++++++++++++++++++++++++++++++++++++++++++++
        """
        node = Path.home()
        home_dir = str(node)
        if isinstance(node, WindowsPath):
            replaced = '%HOMEDRIVE%\\%HOMEPATH%'
        else:
            replaced = '${HOME}'
        new_name = filename.replace(home_dir, replaced)
        return new_name


class Misc:

    message = ''

    @classmethod
    def is_dict(cls, obj):
        return isinstance(obj, typing.Dict)

    @classmethod
    def is_mapping(cls, obj):
        return isinstance(obj, typing.Mapping)

    @classmethod
    def is_list(cls, obj):
        return isinstance(obj, typing.List)

    @classmethod
    def is_mutable_sequence(cls, obj):
        return isinstance(obj, (typing.List, typing.Tuple, typing.Set))

    @classmethod
    def is_sequence(cls, obj):
        return isinstance(obj, typing.Sequence)

    @classmethod
    def try_to_get_number(cls, obj, return_type=None):
        """Try to get a number

        Parameters
        ----------
        obj (object): a number or text number.
        return_type (int, float, bool): a referred return type.

        Returns
        -------
        tuple: status of number and value of number per referred return type
        """
        chk_lst = [int, float, bool]

        if cls.is_string(obj):
            data = obj.strip()
            try:
                if data.lower() == 'true' or data.lower() == 'false':
                    result = True if data.lower() == 'true' else False
                else:
                    result = float(data) if '.' in data else int(data)

                num = return_type(result) if return_type in chk_lst else result
                return True, num
            except Exception as ex:     # noqa
                cls.message = Text(ex)
                return False, obj
        else:
            is_number = cls.is_number(obj)
            num = return_type(obj) if return_type in chk_lst else obj

            if not is_number:
                txt = obj if cls.is_class(obj) else type(obj)
                cls.message = 'Expecting number type, but got {}'.format(txt)
            return is_number, num

    @classmethod
    def is_integer(cls, obj):
        return isinstance(obj, int)

    @classmethod
    def is_boolean(cls, obj):
        return isinstance(obj, bool)

    @classmethod
    def is_float(cls, obj):
        return isinstance(obj, float)

    @classmethod
    def is_number(cls, obj):
        result = cls.is_boolean(obj)
        result |= cls.is_float(obj)
        return result

    @classmethod
    def is_string(cls, obj):
        return isinstance(obj, typing.Text)

    @classmethod
    def is_class(cls, obj):
        return isinstance(obj, typing.Type)     # noqa

    @classmethod
    def is_callable(cls, obj):
        return isinstance(obj, typing.Callable)

    @classmethod
    def is_iterator(cls, obj):
        return isinstance(obj, typing.Iterator)

    @classmethod
    def is_generator(cls, obj):
        return isinstance(obj, typing.Generator)

    @classmethod
    def is_iterable(cls, obj):
        return isinstance(obj, typing.Iterable)

    @classmethod
    def join_string(cls, *args, **kwargs):
        if not args:
            return ''
        if len(args) == 1:
            return str(args[0])

        sep = kwargs.get('separator', '')
        sep = kwargs.get('sep', sep)
        return sep.join(str(item) for item in args)

    @classmethod
    def skip_first_line(cls, data):
        if not cls.is_string(data):
            return data
        else:
            new_data = '\n'.join(data.splitlines()[1:])
            return new_data

    @classmethod
    def substitute_variable(cls, data, root_var_name='self'):
        """substitute variable within data structure if there

        Parameters
        ----------
        data (dict): a dictionary.
        root_var_name (str): root variable of data structure for
                variable substitution.  Default is self.

        Returns
        -------
        dict: a new dictionary if substituted, otherwise, the given data.

        """
        def replace(txt, **kwargs):
            if len(kwargs) == 1:
                var_name = list(kwargs)[0]
                pattern = r'(?i)[{]%s([.][a-z]\w*)+[}]' % var_name
            else:
                pattern = r'(?i)[{][a-z]\w*([.][a-z]\w*)+[}]'
            lines = txt.splitlines()
            for index, line in enumerate(lines):
                if line.strip():
                    lst = []
                    start = 0
                    for match in re.finditer(pattern, line):
                        lst.append(line[start:match.start()])
                        start = match.end()
                        matched_result = match.group()
                        try:
                            val = matched_result.format(**kwargs)
                            lst.append(val)
                        except Exception as ex:     # noqa
                            lst.append(matched_result)
                    else:
                        if lst:
                            lst.append(line[start:])
                            lines[index] = ''.join(lst)

            new_txt = '\n'.join(lines)
            return new_txt

        def substitute(node, variables_):
            if cls.is_dict(node):
                for key, val in node.items():
                    if cls.is_dict(val) or cls.is_list(val):
                        substitute(val, variables_)
                    else:
                        if cls.is_string(val):
                            new_val = replace(val, **variables_)
                            node[key] = new_val
            elif cls.is_list(node):
                for index, item in enumerate(node):
                    if cls.is_dict(item) or cls.is_list(item):
                        substitute(item, variables_)
                    else:
                        if cls.is_string(item):
                            new_item = item.format(obj=variables_)
                            node[index] = new_item
            else:
                return

        if not cls.is_dict(data):
            return data

        new_data = deepcopy(data)
        variables = {root_var_name: DotObject(deepcopy(new_data))}
        substitute(new_data, variables)
        return new_data


class MiscArgs:
    @classmethod
    def get_parsed_result_as_data_or_file(cls, *kwflags, data=''):
        try:
            parser = ArgumentParser(exit_on_error=False)
        except Exception as ex:     # noqa
            parser = ArgumentParser()

        parser.add_argument('val1', nargs='*')
        parser.add_argument('--file', type=str, default='')
        parser.add_argument('--filename', type=str, default='')
        parser.add_argument('--file-name', type=str, default='')
        for flag in kwflags:
            parser.add_argument(flag, type=str, default='')
        parser.add_argument('val2', nargs='*')

        data = str(data).strip()
        first_line = '\n'.join(data.splitlines()[:1])
        pattern = '(?i)file(_?name)?$'

        result = DotObject(
            is_parsed=False, is_data=False, data=data,
            is_file=False, filename='', failure=''
        )

        try:
            options = parser.parse_args(re.split(r' +', first_line))
            result.is_parsed = True
            for flag, val in vars(options).items():
                if re.match(pattern, flag) and val.strip():
                    result.is_file = True
                    result.filename = val.strip()
                else:
                    if flag != 'val1' or flag != 'val2':
                        if val.strip():
                            result.is_data = True
                            result.data = val.strip()
                            return result

            if result.val1 or result.val2:
                result.is_data = True
                return result
            else:
                result.is_parsed = False
                result.failure = 'Invalid data'
                return result

        except Exception as ex:
            result.failure = '{}: {}'.format(type(ex).__name__, ex)
            result.is_parsed = False
            return result


class MiscOutput:
    @classmethod
    def execute_shell_command(cls, cmdline):
        exit_code, output = subprocess.getstatusoutput(cmdline)
        result = DotObject(
            output=output,
            exit_code=exit_code,
            is_success=exit_code == ECODE.SUCCESS
        )
        return result


class DictObject(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.update(*args, **kwargs)

    def __setattr__(self, attr, value):
        super().__setattr__(attr, value)
        self.update(**{attr: value, 'is_updated_attr': False})

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.update({key: value})

    def update(self, *args, is_updated_attr=True, **kwargs):
        obj = dict(*args, **kwargs)
        super().update(obj)
        if is_updated_attr:
            for attr, value in obj.items():
                if Misc.is_string(attr) and re.match(r'(?i)[a-z]\w*$', attr):
                    setattr(self, attr, value)


class DotObject(DictObject):
    def __getattribute__(self, attr):
        value = super().__getattribute__(attr)
        return DotObject(value) if Misc.is_dict(value) else value

    def __getitem__(self, key):
        value = super().__getitem__(key)
        return DotObject(value) if Misc.is_dict(value) else value


def parse_template_result(test_data='', test_file='',
                          template_data='', template_file=''):

    if not any([test_data, test_file]):
        failure = 'Neither test_data nor test_file CANT be empty.'
        raise UtilsParsedTemplateError(failure)
    else:
        if not test_data:
            if File.is_exist(test_file):
                test_data = File.get_content(test_file)
            else:
                failure = '%r test data file is not existed.' % test_file
                raise UtilsParsedTemplateError(failure)

    if not any([template_data, template_file]):
        failure = 'Neither template_data nor template_file CANT be empty.'
        raise UtilsParsedTemplateError(failure)
    else:
        if template_data:
            template = template_data
        else:
            if File.is_exist(template_file):
                template = File.get_content(template_file)
            else:
                failure = '%r template file is not existed.' % template_file
                raise UtilsParsedTemplateError(failure)

    try:
        stream = StringIO(template)
        parser = TextFSM(stream)
        rows = parser.ParseTextToDicts(test_data)

        result = DotObject(
            test_data=test_data, template=template,
            records=rows, records_count=len(rows)
        )
        return result
    except Exception as ex:
        failure = 'BAD-TEMPLATE ({})'.format(Text(ex))
        raise UtilsParsedTemplateError(failure)



"""Top-level module for describe-get-system proof of concept.

- allow end-user to describe his or her requirement to build test script.
"""

from dgspoc.core import Dgs
from dgspoc.config import version

__version__ = version

__all__ = [
    'Dgs',
    'version',
]


"""Module containing the logic for constant definition"""

import re
from enum import IntFlag


class ICSValue:
    """Treating value as ignore case and ignore space during evaluating
    string equality"""
    def __init__(self, value):
        self.value = str(value)

    def __eq__(self, other):
        value1 = self.value.lower()

        if isinstance(other, self.__class__):
            value2 = other.value.lower()
        else:
            value2 = str(other).lower()

        value1 = re.sub(' +', ' ', value1)
        value2 = re.sub(' +', ' ', value2)
        chk = value1.strip() == value2.strip()
        return chk

    def __repr__(self):
        return repr(self.value)

    def __str__(self):
        return self.value


class FrameworkValue:
    def __init__(self, value):
        self.value = str(value)

    def __eq__(self, other):
        value1 = self.value.lower()
        if isinstance(other, self.__class__):
            value2 = other.value.lower()
        else:
            value2 = str(other).lower()

        chk_lst = ['rf', 'robotframework']
        value1, value2 = value1.replace(' ', ''), value2.replace(' ', '')

        if value1 in chk_lst and value2 in chk_lst:
            return True
        else:
            result = value1 == value2
            return result

    def __repr__(self):
        return repr(self.value)

    def __str__(self):
        return self.value


class ECODE(IntFlag):
    SUCCESS = 0
    BAD = 1


class FWTYPE:
    UNITTEST = FrameworkValue('unittest')
    PYTEST = FrameworkValue('pytest')
    ROBOTFRAMEWORK = FrameworkValue('robotframework')


class CONVTYPE:
    CSV = ICSValue('csv')
    JSON = ICSValue('json')
    TEMPLATE = ICSValue('template')



"""Module containing the logic for template storage"""

import yaml
import re
from dgspoc.config import Data
from dgspoc.utils import File
from dgspoc.utils import Misc
from dgspoc.utils import Printer
from dgspoc.utils import Text

from dgspoc.exceptions import TemplateStorageError

from dlapp.utils import convert_wildcard_to_regex


class TemplateStorage:
    message = ''
    filename = Data.template_storage_filename

    @classmethod
    def get(cls, template_id):
        if cls.check(template_id):
            with open(cls.filename) as stream:
                node = yaml.safe_load(stream)
                template = node.get(template_id)
                return template
        else:
            return ''

    @classmethod
    def check(cls, template_id):
        fmt1 = '*** CANT find "{}" template ID because template storage file is empty.'
        fmt2 = '*** CANT find "{}" template ID because template storage file is not created.'
        fmt3 = '{} file has invalid template storage format.'
        if File.is_exist(cls.filename):
            with open(cls.filename) as stream:
                content = stream.read().strip()
                if content:
                    node = yaml.safe_load(content)
                    if Misc.is_dict(node):
                        return template_id in node
                    else:
                        raise TemplateStorageError(fmt3.format(cls.filename))
                else:
                    cls.message = fmt1.format(template_id)
                    return False
        else:
            cls.message = fmt2.format(template_id)
            return False

    @classmethod
    def search(cls, template_id_pattern):
        msg1 = '*** CANT find template ID because template storage file is empty.'
        msg2 = '*** CANT find template ID because template storage file is not created.'
        fmt1 = '{} file has invalid template storage format.'
        fmt2 = '*** There is no template ID matching "{}" pattern.'
        fmt3 = 'Found {} template ID(s) matching "{}" pattern:'
        if File.is_exist(cls.filename):
            with open(cls.filename) as stream:
                content = stream.read().strip()
                if not content:
                    cls.message = Printer.get(msg1)
                    return False

                node = yaml.safe_load(content)

                if not Misc.is_dict(node):
                    raise TemplateStorageError(fmt1.format(cls.filename))

                pattern = convert_wildcard_to_regex(template_id_pattern)
                ids = dict()

                for tmpl_id in sorted(node):
                    if re.search(pattern, tmpl_id, flags=re.I):
                        ids[tmpl_id] = node.get(tmpl_id)

                total = len(ids)

                if total == 0:
                    cls.message = Printer.get(fmt2.format(template_id_pattern))
                    return False

                lst = [fmt3.format(total, template_id_pattern)]
                for tmpl_id in ids:
                    lst.append('  - {}'.format(tmpl_id))

                lst = [Printer.get(lst), '']
                for tmpl_id, template in ids.items():
                    lst.append(Printer.get('Template ID: {}'.format(tmpl_id)))
                    lst.append(template)
                    lst.append('')

                cls.message = '\n'.join(lst)
                return True
        else:
            cls.message = Printer.get(msg2)
            return False

    @classmethod
    def upload(cls, template_id, template, replaced=False):
        try:
            if not File.is_exist(cls.filename):
                File.create(cls.filename)
            if not cls.check(template_id):
                node = File.get_result_from_yaml_file(cls.filename, default=dict())
                node[template_id] = template
                File.save(cls.filename, yaml.safe_dump(node))
                return True
            else:
                if replaced:
                    content = open(cls.filename).read()
                    node = yaml.safe_load(content)
                    node[template_id] = template
                    File.save(cls.filename, yaml.safe_dump(node))
                    return True
                else:
                    fmt = ('CANT upload generated template because of '
                           'duplicate "{}" template ID.  Use replaced '
                           'flag accordingly.')
                    cls.message = fmt.format(template_id)
                    return False
        except Exception as ex:
            cls.message = Text(ex)
            return False

    @classmethod
    def clear(cls, template_id):

        if re.match(r'(?i) *([*]|_+all_+) *$', template_id):
            fmt = 'WONT clear %r because method prohibits multiple clearing template id.'
            cls.message = fmt % template_id
            return False

        yaml_obj = File.get_result_from_yaml_file(cls.filename)
        if template_id in yaml_obj:
            yaml_obj.pop(template_id)
            File.save(cls.filename, yaml.safe_dump(yaml_obj))
            return True
        else:
            fmt = 'CANT clear %r because it is not in storage template.'
            cls.message = fmt % template_id
            return False



"""Module containing the logic for console command line examples"""

from yaml import safe_load
from pathlib import Path
from dgspoc.utils import Printer


def get_number_of_example(name, default=1):
    file_obj = Path(Path(__file__).parent, 'exampledata.yaml')
    content = file_obj.read_text()
    dict_obj = safe_load(content)
    result = dict_obj.get(name, dict())
    total = len(result) or default
    return total


class Example:
    name = 'example'

    @classmethod
    def get(cls, index):
        file_obj = Path(Path(__file__).parent, 'exampledata.yaml')
        content = file_obj.read_text()
        dict_obj = safe_load(content)
        node = dict_obj.get(cls.name).get('example{}'.format(index))
        header = node.get('header')
        header = Printer.get(header)
        body = node.get('body')
        example_text = '{}\n\n{}'.format(header, body)
        return example_text


class BuildTemplateExample(Example):
    name = 'build_template'


class SearchTemplateExample(Example):
    name = 'search_template'


class TestTemplateExample(Example):
    name = 'test_template'



"""Module containing the exception class for describe-get-system proof of concept module."""


class DGSError(Exception):
    """Use to capture error DGS construction."""


class TemplateStorageError(DGSError):
    """Use to capture error for TemplateStorage."""


class AdaptorAuthenticationError(DGSError):
    """Use to capture error for adaptor connection"""


class InterpreterError(DGSError):
    """Use to capture error for interpreter"""


class NotImplementedFrameworkError(InterpreterError):
    """Use to capture error for not implement framework"""


class ComparisonOperatorError(InterpreterError):
    """Use to capture error for invalid comparison operator"""


class ConnectDeviceStatementError(InterpreterError):
    """Use to capture error for interpreting connect device statement"""


class DisconnectDeviceStatementError(InterpreterError):
    """Use to capture error for interpreting disconnect device statement"""


class ReleaseDeviceStatementError(InterpreterError):
    """Use to capture error for interpreting release device statement"""


class WaitForStatementError(InterpreterError):
    """Use to capture error for interpreting wait for statement"""


class PerformerStatementError(InterpreterError):
    """Use to capture error for performer statement"""


class VerificationStatementError(InterpreterError):
    """Use to capture error for verification statement"""


class ScriptBuilderError(InterpreterError):
    """Use to capture error for building test script"""


class DurationArgumentError(InterpreterError):
    """Use to capture error for wait_for function"""


class ConvertorTypeError(InterpreterError):
    """Use to capture convertor argument for filter method"""


class TemplateReferenceError(InterpreterError):
    """Use to capture template reference for filter method"""


class UtilsParsedTemplateError(InterpreterError):
    """Use to capture error for parsing template in utils"""



"""Module containing the logic for describe-get-system to parse
user describing request"""

import re

from dgspoc.utils import DotObject


class CheckStatement:

    @classmethod
    def is_execute_cmdline(cls, data):
        pattern = r'(?i) *[{]?([a-z0-9]\S*)(, ?([a-z0-9]\S*))*[}]? +exec(utes?)? +'
        match = re.match(pattern, data)
        return bool(match)

    @classmethod
    def is_performer_statement(cls, data):
        pattern = (r'(?i) *[{]?([a-z0-9]\S*)(, ?([a-z0-9]\S*))*[}]? +'
                   r'(exec(utes?)?|conf(ig(ures?)?)?|reloads?) +')
        match = re.match(pattern, data)
        return bool(match)

    @classmethod
    def is_verification_statement(cls, data):
        pattern = r'(?i) +must +be +\S+( +[0-9]+)? *$'
        chk = bool(re.search(pattern, data))
        chk &= cls.is_execute_cmdline(data)
        return chk

    @classmethod
    def is_regular_iterative_statement(cls, data):
        pattern = r'(?i) *loop +[0-9]+ +times? *$'
        match = re.match(pattern, data)
        return bool(match)

    @classmethod
    def is_until_iterative_statement(cls, data):
        pattern = r'(?i) *loop +[0-9]+ +times? +until *$'
        match = re.match(pattern, data)
        return bool(match)

    @classmethod
    def is_to_last_iterative_statement(cls, data):
        pattern = r'(?i) *loop +[0-9]+ +times? +to +last *$'
        match = re.match(pattern, data)
        return bool(match)

    @classmethod
    def is_iterative_statement(cls, data):
        chk = cls.is_regular_iterative_statement(data)
        chk |= cls.is_until_iterative_statement(data)
        chk |= cls.is_to_last_iterative_statement(data)
        return chk

    @classmethod
    def is_pausing_statement(cls, data):
        match = re.match(r'(?i) *(sleeps?|(wait +for)) +', data)
        return bool(match)

    @classmethod
    def is_connect_device_statement(cls, data):
        match = re.match(r'(?i) *connects? +device +', data)
        return bool(match)

    @classmethod
    def is_disconnect_device_statement(cls, data):
        match = re.match(r'(?i) *disconnects?( +device)? +', data)
        return bool(match)

    @classmethod
    def is_release_device_statement(cls, data):
        match = re.match(r'(?i) *releases? +device +', data)
        return bool(match)

    @classmethod
    def is_dummy_statement(cls, data):
        match = re.match(r'(?i) *dummy[_. -]*(pass|fail)', data)
        return bool(match)

    @classmethod
    def is_child_execute_cmdline(cls, data):
        chk = cls.is_execute_cmdline(data) and data.startswith(' ')
        return chk

    @classmethod
    def is_child_performer_statement(cls, data):
        chk = cls.is_performer_statement(data) and data.startswith(' ')
        return chk

    @classmethod
    def is_child_verification_statement(cls, data):
        chk = cls.is_verification_statement(data) and data.startswith(' ')
        return chk

    @classmethod
    def is_child_iterative_statement(cls, data):
        chk = cls.is_iterative_statement(data) and data.startswith(' ')
        return chk

    @classmethod
    def is_child_pausing_statement(cls, data):
        chk = cls.is_pausing_statement(data) and data.startswith(' ')
        return chk

    @classmethod
    def is_child_connect_device_statement(cls, data):
        chk = cls.is_connect_device_statement(data) and data.startswith(' ')
        return chk

    @classmethod
    def is_child_disconnect_device_statement(cls, data):
        chk = cls.is_disconnect_device_statement(data) and data.startswith(' ')
        return chk

    @classmethod
    def is_child_release_device_statement(cls, data):
        chk = cls.is_release_device_statement(data) and data.startswith(' ')
        return chk

    @classmethod
    def is_child_dummy_statement(cls, data):
        chk = cls.is_dummy_statement(data) and data.startswith(' ')
        return chk


class ParsedOperation:
    def __init__(self, data):
        self.name = ''
        self.devices_names = []
        self.operation_type = ''
        self.data = data
        self.operation_ref = ''
        self.convertor = ''
        self.convertor_arg = ''
        self._has_select_statement = False
        self.select_statement = ''
        self.condition = ''
        self.expected_condition = -1
        self.condition_data = ''

        self.error = ''

        self.parse()

    @property
    def is_parsed(self):
        return self.error == ''

    @property
    def is_execution(self):
        return self.operation_type == 'execution'

    @property
    def is_configuration(self):
        return self.operation_type == 'configuration'

    @property
    def is_verification(self):
        chk = self.is_execution
        chk &= self.convertor != ''
        chk &= self.select_statement != ''
        chk &= self.condition != ''
        return chk

    @property
    def is_reload(self):
        return self.operation_type == 'reload'

    @property
    def is_valid_operation(self):
        result = self.is_execution
        result |= self.is_configuration
        result |= self.is_reload
        return result

    @property
    def has_select_statement(self):
        return self._has_select_statement

    @property
    def is_need_verification(self):
        return self.condition != '' and self.expected_condition >= 0

    @property
    def is_json(self):
        return self.convertor == 'json'

    @property
    def is_csv(self):
        return self.convertor == 'csv'

    @property
    def is_template(self):
        return self.convertor == 'template'

    def parse(self):
        pattern = (r'(?i) *(?P<devices_names>[{]?([a-z0-9]\S*)(, ?([a-z0-9]\S*))*[}]?) +'
                   r'(?P<operation_type>exec(utes?)?|conf(ig(ures?)?)?|reloads?) +'
                   r'(?P<operation>.+) *$')
        match = re.match(pattern, self.data)
        if not match:
            return

        node = DotObject(match.groupdict())

        self.parse_devices_names(node.devices_names)
        self.is_parsed and self.parse_operation_type(node.operation_type)   # noqa

        if self.is_parsed and self.is_valid_operation:
            if self.is_execution:
                result = ExecuteOperation(node.operation)
                result.sync(other=self)
            elif self.is_configuration:
                result = ConfigOperation(node.operation)
                result.sync(other=self)
            elif self.is_reload:
                result = ReloadOperation(node.operation)
                result.sync(other=self)

    def parse_devices_names(self, data):
        other_data = data.strip().lstrip('{').rstrip('}')
        pattern = r'([a-z0-9]\S*)(, ?([a-z0-9]\S*))*$'
        match = re.match(pattern, other_data, re.I)
        if match:
            names = re.split(r' *, *', other_data)
            if names:
                self.devices_names = names
                return
        self.error = 'Invalid devices names format ("%s")' % data

    def parse_operation_type(self, data):
        tbl = dict(e='execution', c='configuration', r='reload')
        key = data.strip().lower()[:1]
        self.operation_type = tbl.get(key, '')
        if not self.operation_type:
            fmt = 'Invalid operation ("%s").  It MUST BE execute, configure, or reload.'
            self.error = fmt % data


class ExecuteOperation:
    def __init__(self, data):
        self.name = 'execution'
        self._remaining_data = ''
        self.data = data
        self.operation_ref = ''
        self.convertor = ''
        self.convertor_arg = ''
        self.select_statement = ''
        self.condition = ''
        self.expected_condition = -1
        self.condition_data = ''

        self.error = ''

        self.parse()

    @property
    def is_parsed(self):
        return self.error == ''

    @property
    def has_select_statement(self):
        return self.select_statement != ''

    @property
    def is_need_verification(self):
        return self.condition != '' and self.expected_condition >= 0

    def parse(self):
        self._remaining_data = self.data
        self.parse_condition()
        self.parse_select_statement()
        self.parse_convertor()
        self.operation_ref = self._remaining_data

    def parse_condition(self):

        data = self._remaining_data

        pattern = r'(?i) +must +be +\S+( +\S+)? *$'
        match = re.search(pattern, data)
        if not match:
            return

        verified_str = match.group().strip()

        pattern = (r'(?i)must +be +((?P<val1>true|false)|'
                   r'((?P<op>\S+) +(?P<val2>[0-9]+)))$')
        match = re.search(pattern, verified_str)

        if not match:
            fmt = 'Invalid command line verification format (Unexpected: %s)'
            self.error = fmt % verified_str
            return

        self.condition_data = match.group().strip()
        self._remaining_data = re.sub(pattern, '', data)

        node = DotObject(match.groupdict())
        if node.val1:
            self.condition = 'eq'
            self.expected_condition = 1 if node.val1.lower() == 'true' else 0
        else:
            tbl = dict(
                EQ='eq', EQUAL='eq', EQUAL_TO='eq',
                NE='ne', NOT_EQUAL='ne', NOT_EQUAL_TO='ne',
                GT='gt', GREATER_THAN='gt',
                GE='ge', GREATER_THAN_OR_EQUAL='ge', EQUAL_OR_GREATER_THAN='ge',
                GREATER_THAN_OR_EQUAL_TO='ge', EQUAL_TO_OR_GREATER_THAN='ge',
                LT='lt', LESS_THAN='lt',
                LE='lt', LESS_THAN_OR_EQUAL='lt', EQUAL_OR_LESS_THAN='lt',
                LESS_THAN_OR_EQUAL_TO='lt', EQUAL_TO_OR_LESS_THAN='lt'
            )
            tbl.update({'==': 'eq', '!=': 'ne', '>': 'gt',
                        '>=': 'ge', '<': 'lt', '<=': 'le'})

            op = node.op.upper()
            found = [True for k, v in tbl.items() if k == op or v == op]
            if any(found):
                self.condition = tbl[op] if op in tbl else op
                self.expected_condition = int(node.val2)
            else:
                fmt = 'Invalid comparison operator (%r).  It must be %s'
                self.error = fmt % (node.op, list(tbl))

    def parse_select_statement(self):

        data = self._remaining_data.strip()

        pattern = r'(?i) +(?P<sel_stmt>select +([*]|_+all_+|[a-z].+) where +.+)'
        match = re.search(pattern, data)
        if not match:
            return

        self.select_statement = match.group('sel_stmt').strip()
        self._remaining_data = re.sub(pattern, '', data)

    def parse_convertor(self):

        data = self._remaining_data.strip()

        pattern = r'(?i) using[- _]+(?P<type>csv|json|template)([ =]+(?P<arg>\S+))? *$'
        match = re.search(pattern, data)
        if not match:
            if self.has_select_statement:
                self.error = 'Invalid command line verification without convertor'
            return
        self.convertor = match.group('type').lower()
        if match.group('arg'):
            self.convertor_arg = match.group('arg').strip().strip('=')
        self._remaining_data = re.sub(pattern, '', data)

    def sync(self, other=None):
        if isinstance(other, ParsedOperation):
            other.name = self.name
            other.operation_ref = self.operation_ref
            other.convertor = self.convertor
            other.convertor_arg = self.convertor_arg
            other.select_statement = self.select_statement
            other.condition = self.condition
            other.expected_condition = self.expected_condition
            other.condition_data = self.condition_data
            other._has_select_statement = self.has_select_statement
            other.error = self.error


class ConfigOperation:
    def __init__(self, data):
        self.name = 'configuration'
        self.operation_ref = ''
        self.data = data

        self.error = ''

        self.parse()

    @property
    def is_parsed(self):
        return self.error == ''

    def parse(self):
        self.operation_ref = self.data

    def sync(self, other=None):
        if isinstance(other, ParsedOperation):
            other.name = self.name
            other.operation_ref = self.operation_ref


class ReloadOperation:
    def __init__(self, data):
        self.name = 'reload'
        self.operation_ref = ''
        self.data = data

        self.error = ''

        self.parse()

    @property
    def is_parsed(self):
        return self.error == ''

    def parse(self):
        self.operation_ref = self.data

    def sync(self, other=None):
        if isinstance(other, ParsedOperation):
            other.name = self.name
            other.operation_ref = self.operation_ref



"""Module containing logic for describe-get-system proof of concept."""

import time
from io import StringIO

from textfsm import TextFSM


from dlapp import create_from_csv_data
from dlapp import create_from_json_data
from dlapp import DLQuery

from dgspoc.adaptor import Adaptor

from dgspoc.constant import CONVTYPE

from dgspoc.storage import TemplateStorage

from dgspoc.utils import DotObject
from dgspoc.utils import File
from dgspoc.utils import Misc
from dgspoc.utils import MiscArgs

from dgspoc.exceptions import DurationArgumentError
from dgspoc.exceptions import ConvertorTypeError
from dgspoc.exceptions import TemplateReferenceError


class Dgs:
    """Describe-Get-System class"""
    kwargs = DotObject()

    @classmethod
    def wait_for(cls, duration):
        """pausing method

        Parameters
        ----------
        duration (float): total seconds
        """

        is_number, number = Misc.try_to_get_number(duration, return_type=float)
        if is_number:
            time.sleep(number)
        else:
            failure = 'duration must be a number (unexpected: %s)' % duration
            raise DurationArgumentError(failure)

    @classmethod
    def connect_device(cls, host, adaptor='unreal-device', **kwargs):
        """generic connecting device

        Parameters
        ----------
        host (str): address of device
        adaptor (str): connection adaptor.
        kwargs (dict): additional keyword arguments for connecting device

        Returns
        -------
        Adaptor: a device connection
        """

        connection = Adaptor(adaptor, host, **kwargs)
        connection.connect()
        return connection

    @classmethod
    def disconnect_device(cls, connection, **kwargs):
        """generic disconnecting device

        Parameters
        ----------
        connection (Adaptor): instance of device connection
        kwargs (dict): additional keyword arguments for disconnection device connection

        Returns
        -------
        bool: True if successfully disconnected device connection, otherwise, False.
        """
        result = connection.disconnect(**kwargs)
        return result

    @classmethod
    def release_device(cls, connection, **kwargs):
        """generic releasing device

        Parameters
        ----------
        connection (Adaptor): instance of device connection
        kwargs (dict): additional keyword arguments for releasing device connection

        Returns
        -------
        bool: True if successfully released device connection, otherwise, False.
        """
        result = connection.release(**kwargs)
        return result

    @classmethod
    def execute_cmdline(cls, connection, cmdline, **kwargs):
        """generic device execution

        Parameters
        ----------
        connection (Adaptor): instance of device connection
        cmdline (str): command lines
        kwargs (dict): additional keyword arguments for command line execution

        Returns
        -------
        str: output of command line
        """
        result = connection.execute(cmdline, **kwargs)
        return result

    @classmethod
    def configure_device(cls, connection, cfg, **kwargs):
        """generic device configuration

        Parameters
        ----------
        connection (Adaptor): instance of device connection
        cfg (str): configuration lines
        kwargs (dict): additional keyword arguments for configuring device

        Returns
        -------
        str: the result of device configuration
        """
        result = connection.configure(cfg, **kwargs)
        return result

    @classmethod
    def reload_device(cls, connection, reload_command, **kwargs):
        """generic reloading device

        Parameters
        ----------
        connection (Adaptor): instance of device connection
        reload_command (str): reload command
        kwargs (dict): additional keyword arguments for reloading device

        Returns
        -------
        str: the result of reloading process
        """
        result = connection.reload(reload_command, **kwargs)
        return result

    @classmethod
    def convert_and_filter(cls, text, convertor='', template_ref='', select_statement=''):
        """generic method to convert text to data struct and filter result
        per select_statement

        Parameters
        ----------
        text (str): output or text data
        convertor (str): cvs, json, or template
        template_ref (str): template-id or template filename
        select_statement (str): a select statement

        Returns
        -------
        list: the list of records
        """

        if convertor == CONVTYPE.CSV:
            result = cls.do_filter_csv(text, select_statement=select_statement)
            return result

        elif convertor == CONVTYPE.JSON:
            result = cls.do_filter_json(text, select_statement=select_statement)
            return result
        elif convertor == CONVTYPE.TEMPLATE:
            result = cls.do_filter_template(text, template_ref,
                                            select_statement=select_statement)
            return result
        else:
            fmt = 'convertor MUST BE csv, json, or template (Unexpected: %s)'
            raise ConvertorTypeError(fmt % convertor)

    @classmethod
    def do_filter_csv(cls, text, select_statement=''):
        """generic method to convert csv text to list of dict and filter
        per select_statement

        Parameters
        ----------
        text (str): output or text data
        select_statement (str): a select statement

        Returns
        -------
        list: the list of records
        """

        query_obj = create_from_csv_data(text)
        result = query_obj.find(select_statement=select_statement)
        return result

    @classmethod
    def do_filter_json(cls, text, select_statement=''):
        """generic method to convert json text to data structure and filter
        per select_statement

        Parameters
        ----------
        text (str): output or text data
        select_statement (str): a select statement

        Returns
        -------
        list: the list of records
        """
        query_obj = create_from_json_data(text)
        result = query_obj.find(select_statement=select_statement)
        return result

    @classmethod
    def do_filter_template(cls, text, tmpl_ref, select_statement=''):
        """generic method to convert text to data struct using TextFSM and
        filter per select_statement

        Parameters
        ----------
        text (str): output or text data
        template_ref (str): template-id or template filename
        select_statement (str): a select statement

        Returns
        -------
        list: the list of records
        """

        tmpl_ref = str(tmpl_ref).strip()

        if not tmpl_ref:
            failure = 'Template reference CANT BE empty.'
            raise TemplateReferenceError(failure)

        parsed_args_result = MiscArgs.get_parsed_result_as_data_or_file('--template-id')

        if not parsed_args_result.is_parsed:
            fmt = 'Invalid template reference format (Unexpected: %r)'
            raise TemplateReferenceError(fmt % tmpl_ref)

        if parsed_args_result.is_file:
            fn = parsed_args_result.filename
            if not File.is_exist(fn):
                fmt = '%r template file CANT BE FOUND.'
                raise TemplateReferenceError(fmt % fn)
            else:
                with open(fn) as stream:
                    template_parser = TextFSM(stream)
        else:
            tmpl_id = parsed_args_result.data
            if not TemplateStorage.check(tmpl_id):
                fmt = '%r template-id CANT BE FOUND in template storage.'
                raise TemplateReferenceError(fmt % tmpl_id)
            else:
                template = TemplateStorage.get(tmpl_id)
                stream = StringIO(template)
                template_parser = TextFSM(stream)

        rows = template_parser.ParseTextToDicts(text)

        if not select_statement:
            return rows
        else:
            query_node = DLQuery(rows)
            result = query_node.find(select=select_statement)
            return result



"""Module containing the attributes for describe-get-system proof of concept module."""

from os import path
from textwrap import dedent
import yaml

import templateapp
import regexapp
import dlapp

import pytest
import robot

from dgspoc.utils import File
from dgspoc.utils import Misc

__version__ = '0.3.3'
version = __version__

__all__ = [
    'version',
    'Data'
]


class Data:

    # app yaml files
    app_directory = File.get_path('.geekstrident', 'dgspoc', is_home=True)
    template_storage_filename = File.get_path(app_directory, 'template_storage.yaml')

    # main app
    main_app_text = 'dgs v{}'.format(version)

    # company
    company = 'Geeks Trident LLC'
    company_url = 'https://www.geekstrident.com/'

    # URL
    repo_url = 'https://github.com/Geeks-Trident-LLC/dgspoc'
    # TODO: Need to update wiki page for documentation_url instead of README.md.
    documentation_url = path.join(repo_url, 'blob/develop/README.md')
    license_url = path.join(repo_url, 'blob/develop/LICENSE')

    # License
    years = '2022-2040'
    license_name = 'BSD 3-Clause License'
    copyright_text = 'Copyright @ {}'.format(years)
    license = dedent(
        """
        BSD 3-Clause License

        Copyright (c) {}, {}
        All rights reserved.

        Redistribution and use in source and binary forms, with or without
        modification, are permitted provided that the following conditions are met:

        1. Redistributions of source code must retain the above copyright notice, this
           list of conditions and the following disclaimer.

        2. Redistributions in binary form must reproduce the above copyright notice,
           this list of conditions and the following disclaimer in the documentation
           and/or other materials provided with the distribution.

        3. Neither the name of the copyright holder nor the names of its
           contributors may be used to endorse or promote products derived from
           this software without specific prior written permission.

        THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
        AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
        IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
        DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
        FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
        DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
        SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
        CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
        OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
        OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
        """.format(years, company)
    ).strip()

    @classmethod
    def get_app_info(cls):
        from platform import uname as u, python_version as v
        lst = [cls.main_app_text,
               'Project : {}'.format(cls.repo_url),
               'License : {}'.format(cls.license_name),
               'Platform: {0.system} {0.release} - Python {1}'.format(u(), v()),
               ]
        app_info = '\n'.join(lst)
        return app_info

    @classmethod
    def get_dependency(cls):
        obj = dict(
            templateapp=dict(
                package='templateapp v{}'.format(templateapp.version),
                url='https://pypi.org/project/templateapp/'
            ),
            pytest=dict(
                package='pytest v{}'.format(pytest.__version__),
                url='https://pypi.org/project/pytest/'
            ),
            robotframework=dict(
                package='robotframework v{}'.format(robot.__version__),
                url='https://pypi.org/project/robotframework/'
            ),
        )
        obj.update(templateapp.config.Data.get_dependency())
        obj.update(regexapp.config.Data.get_dependency())
        obj.update(dlapp.config.Data.get_dependency())
        dependencies = dict(sorted(obj.items(), key=lambda x: str(x[0])))   # noqa
        return dependencies

    @classmethod
    def get_template_storage_info(cls):
        fn = cls.template_storage_filename
        generic_fn = File.change_home_dir_to_generic(fn)
        if File.is_exist(fn):
            existed = 'Yes'
            with open(fn) as stream:
                node = yaml.safe_load(stream)
                total = len(node) if Misc.is_dict(node) else 0
        else:
            existed = 'No'
            total = 0
        lst = [
            'Template Storage Info:',
            # '  - Location: {}'.format(fn),
            '  - Location: {}'.format(generic_fn),
            '  - Existed: {}'.format(existed),
            '  - Total Templates: {}'.format(total)
        ]
        return '\n'.join(lst)



"""Module containing the logic for describe-get-system operation"""

import sys
import re
import argparse
import operator

from functools import partial
from textwrap import wrap

from dgspoc.utils import File
from dgspoc.utils import Printer
from dgspoc.utils import Text
from dgspoc.utils import DotObject
from dgspoc.utils import parse_template_result
from dgspoc.utils import Misc
from dgspoc.utils import MiscArgs

from dgspoc.constant import ECODE
from dgspoc.constant import CONVTYPE

from dgspoc.storage import TemplateStorage

from dgspoc.usage import validate_usage
from dgspoc.usage import validate_example_usage
from dgspoc.usage import show_usage

from dgspoc.adaptor import Adaptor

from dgspoc.parser import ParsedOperation
from dgspoc.parser import CheckStatement

from templateapp import TemplateBuilder

from dlapp import DLQuery
from dlapp import create_from_csv_data
from dlapp import create_from_json_data

from pprint import pprint
from dlapp.collection import Tabular


def do_build_template(options):
    command, operands = options.command, list(options.operands)
    op_count = len(operands)
    feature = str(operands[0]).lower().strip() if op_count > 0 else ''
    if command == 'build' and feature == 'template':
        operands = operands[1:]
        name = '{}_{}'.format(command, feature)
        validate_usage(name, operands)
        validate_example_usage(name, operands)

        op_txt = ' '.join(operands).rstrip()

        if not op_txt:
            show_usage(name, exit_code=ECODE.BAD)

        if File.is_exist(op_txt):
            with open(op_txt) as stream:
                user_data = stream.read()
        else:
            user_data = op_txt

        try:
            factory = TemplateBuilder(
                user_data=user_data, author=options.author, email=options.email,
                company=options.company
            )

            template_id = options.tmplid.strip()
            filename = options.filename.strip()

            fmt1 = '+++ Successfully uploaded generated template to "{}" template ID.'
            fmt2 = '+++ Successfully saved generated template to {}'
            fmt3 = 'CANT save generated template to existing {} file.  Use replaced flag accordingly.'

            if template_id or filename:
                is_ok = True
                lst = []
                if template_id:
                    is_uploaded = TemplateStorage.upload(
                        template_id, factory.template, replaced=options.replaced
                    )
                    is_ok &= is_uploaded
                    msg = fmt1.format(template_id) if is_uploaded else TemplateStorage.message
                    lst.append(msg)
                if filename:
                    filename = File.get_path(filename)
                    if File.is_exist(filename) and not options.replaced:
                        msg = fmt3.format(filename)
                        is_ok &= False
                    else:
                        is_saved = File.save(options.filename, factory.template)
                        is_ok &= is_saved
                        msg = fmt2.format(filename) if is_saved else File.message

                    lst and lst.append('=' * 20)
                    lst.append(msg)

                lst and Printer.print(lst)
                sys.exit(ECODE.SUCCESS if is_ok else ECODE.BAD)
            else:
                print(factory.template)
                sys.exit(ECODE.SUCCESS)

        except Exception as ex:
            print(Text(ex))
            sys.exit(ECODE.BAD)

    elif command == 'build' and feature != 'template':
        if feature == 'script':
            return
        else:
            exit_code = ECODE.SUCCESS if feature == 'usage' else ECODE.BAD
            show_usage(command, exit_code=exit_code)


def do_search_template(options):
    command, operands = options.command, list(options.operands)
    op_count = len(operands)
    feature = str(operands[0]).lower().strip() if op_count > 0 else ''
    if command == 'search' and feature == 'template':
        operands = operands[1:]
        name = '{}_{}'.format(command, feature)
        validate_usage(name, operands)
        validate_example_usage(name, operands)

        op_txt = ' '.join(operands).rstrip()

        if not op_txt:
            show_usage(name, exit_code=ECODE.BAD)

        tmpl_id_pattern = operands[0]
        is_found = TemplateStorage.search(tmpl_id_pattern)
        print(TemplateStorage.message)
        sys.exit(ECODE.SUCCESS if is_found else ECODE.BAD)

    elif command == 'search' and feature != 'template':
        exit_code = ECODE.SUCCESS if feature == 'usage' else ECODE.BAD
        show_usage('{}_template'.format(command), exit_code=exit_code)


def do_clear_template(options):
    template_id = options.template_id.strip()

    if re.match(r'(?i) *([*]|_+all_+) *$', template_id):
        fmt = ('*** Failed to clear %r template reference because '
               'Describe-Get-System prohibits multiple clearing')
        print(fmt % template_id)
        sys.exit(ECODE.BAD)

    if template_id:
        result = MiscArgs.get_parsed_result_as_data_or_file(data=template_id)
        if result.is_file:
            is_deleted = File.delete(result.filename)
            print('%s %s' % ('+++' if is_deleted else '***', File.message))
            sys.exit(ECODE.SUCCESS if is_deleted else ECODE.BAD)
        else:
            is_cleared = TemplateStorage.clear(template_id)
            if is_cleared:
                fmt = '+++ Successfully cleared %r template id in template storage.'
                message = fmt % template_id
            else:
                message = '*** %s' % TemplateStorage.message
            print(message)
            sys.exit(ECODE.SUCCESS if is_cleared else ECODE.BAD)


def validate_test_data_flag(options):
    command = options.command
    feature = options.operands[0] if options.operands else ''
    if options.testfile == '' and options.adaptor == '':
        lst = ['CANT run {} test WITHOUT test data.'.format(feature),
               'Please use --test-file=<test-file-name> or',
               '           --adaptor=<adaptor_name> --execution="<device cmdline>"']
        Printer.print(lst)
        show_usage('{}_{}'.format(command, feature), exit_code=ECODE.BAD)


def get_test_from_adaptor(options):
    command = options.command
    feature = options.operands[0] if options.operands else ''
    name = '{}_{}'.format(command, feature)
    execution = options.execution.strip()
    if not execution:
        lst = [
            'ExecutionSyntaxError: must be',
            '--execution="--host=<addr_or_name> <cmdline>"',
        ]
        Printer.print(lst)
        show_usage(name, exit_code=ECODE.BAD)

    try:

        lst = execution.split(' ')

        parser = argparse.ArgumentParser(exit_on_error=False)
        parser.add_argument('items', nargs='*')
        parser.add_argument('--host', type=str, default='')
        parser.add_argument('other_items', nargs='*')
        parser_args = parser.parse_args(lst)

        host = parser_args.host
        cmdline = ' '.join(parser_args.items + parser_args.other_items)

        if not parser_args.host:
            lst = [
                'ExecutionSyntaxError: must be',
                '--execution="--host=<addr_or_name> <cmdline>"',
            ]
            Printer.print(lst)
            show_usage(name, exit_code=ECODE.BAD)

        device = Adaptor(options.adaptor, host)
        device.connect()
        test_data = device.execute(cmdline)
        device.disconnect()
        device.release()
        return test_data
    except Exception as ex:
        failure = 'AdaptorInquiryError - ({})'.format(Text(ex))
        Printer.print(failure)
        sys.exit(ECODE.BAD)


def do_testing(options):
    pattern = '^--(template-id|(file(-?name)?))='
    command, operands = options.command, list(options.operands)
    adaptor = options.adaptor.strip().lower()
    action = options.action.strip()
    if command == 'test':
        name = command
        validate_usage(name, operands)
        validate_example_usage(name, operands)
        is_showed = False
        if re.search(r'(?i)--showed\b', action):
            is_showed = True
            action = re.sub(r'(?i)--showed\b', '', action).strip()

        is_tabular = False
        if re.search(r'(?i)--tabular\b', action):
            is_tabular = True
            action = re.sub(r'(?i)--tabular\b', '', action).strip()

        adaptor = adaptor or 'stream'
        is_adaptor_stream = bool(re.match(r'(stream|file(name)?)$', adaptor))
        is_execute_cmdline = CheckStatement.is_execute_cmdline(action)

        if is_adaptor_stream:
            data = 'dummy execute %s' % action if not is_execute_cmdline else action

            node = ParsedOperation(data)
            test_data_file = re.sub(pattern, '', node.operation_ref.strip())
            File.message = ''
            output = File.get_content(test_data_file)
            if File.message:
                print('*** %s' % File.message)
                sys.exit(ECODE.BAD)
            print(output)
            result = DotObject(test_data=output, template='', records=[], records_count=0)

            if node.has_select_statement or node.convertor:
                if node.is_csv or node.is_json:
                    try:
                        lines = result.test_data.splitlines()
                        index = 0
                        pat = (r'(?i)\w{3} +\d\d? +\d{4} '
                               r'\d\d:\d\d:\d\d[.]\d\d\d for "\S+" - '
                               r'UNREAL-DEVICE-\w+-SERVICE-TIMESTAMP')
                        for i, line in enumerate(lines):
                            if re.match(pat, line):
                                index = i
                                break
                        test_data = '\n'.join(lines[index+1:])
                        method = create_from_json_data if node.is_json else create_from_csv_data
                        records = method(test_data).data
                        result.records = records
                        result.records_count = len(records)
                    except Exception as ex:
                        print('*** %s' % Text(ex))
                        sys.exit(ECODE.BAD)
                else:
                    pfunc = partial(parse_template_result, test_file=test_data_file)
                    if not node.convertor == CONVTYPE.TEMPLATE or not node.convertor_arg:
                        print('*** Invalid action: %s ***' % action)
                        sys.exit(ECODE.BAD)

                    try:
                        if re.match('(?i)--file', node.convertor_arg.strip()):
                            template_file = re.sub(pattern, '', node.convertor_arg.strip())
                            result = pfunc(template_file=template_file)
                        else:
                            tmpl_id = re.sub(pattern, '', node.convertor_arg.strip())
                            if TemplateStorage.check(tmpl_id):
                                tmpl_data = TemplateStorage.get(tmpl_id)
                                result = pfunc(template_data=tmpl_data)
                            else:
                                fmt = '*** %r template id CANT find in template storage.'
                                print(fmt % tmpl_id)
                                sys.exit(ECODE.BAD)
                    except Exception as ex:
                        failure = '*** %s' % ex
                        print(failure)
                        sys.exit(ECODE.BAD)
            else:
                sys.exit(ECODE.SUCCESS)
        else:
            if not CheckStatement.is_performer_statement(action):
                failure = '*** Invalid action: %s ***' % action
                print(failure)
                sys.exit(ECODE.BAD)

            node = ParsedOperation(action)
            if not node.devices_names:
                failure = '*** Invalid action: %s ***' % action
                print(failure)
                sys.exit(ECODE.BAD)

            host = node.devices_names[0]
            connection = Adaptor(adaptor, host)
            tbl = dict(execution=connection.execute,
                       configuration=connection.configure,
                       reload=connection.reload)
            operation_method = tbl[node.name]
            connection.connect()
            output = operation_method(node.operation_ref)
            connection.release()

            if node.is_reload or node.is_configuration:
                print(output)
                sys.exit(ECODE.SUCCESS)

            if node.is_csv or node.is_json:
                try:
                    lines = output.test_data.splitlines()
                    index = 0
                    pat = (r'(?i)\w{3} +\d\d? +\d{4} '
                           r'\d\d:\d\d:\d\d[.]\d\d\d for "\S+" - '
                           r'UNREAL-DEVICE-\w+-SERVICE-TIMESTAMP')
                    for i, line in enumerate(lines):
                        if re.match(pat, line):
                            index = i
                            break
                    test_data = '\n'.join(lines[index + 1:])
                    method = create_from_json_data if node.is_json else create_from_csv_data
                    records = method(test_data).data
                    result = DotObject(
                        test_data=output, template='',
                        records=records, records_count=len(records)
                    )
                except Exception as ex:
                    print('*** %s' % Text(ex))
                    sys.exit(ECODE.BAD)
            elif node.is_template:
                pfunc = partial(parse_template_result, test_data=output)
                try:
                    if re.match('(?i)--file', node.convertor_arg.strip()):
                        template_file = re.sub(pattern, '', node.convertor_arg.strip())
                        result = pfunc(template_file=template_file)
                    else:
                        tmpl_id = re.sub(pattern, '', node.convertor_arg.strip())
                        if TemplateStorage.check(tmpl_id):
                            tmpl_data = TemplateStorage.get(tmpl_id)
                            result = pfunc(template_data=tmpl_data)
                        else:
                            fmt = '*** %r template id CANT find in template storage.'
                            print(fmt % tmpl_id)
                            sys.exit(ECODE.BAD)
                except Exception as ex:
                    failure = '*** %s' % ex
                    print(failure)
                    sys.exit(ECODE.BAD)
            else:
                if not node.has_select_statement:
                    sys.exit(ECODE.SUCCESS)
                else:
                    failure = '*** Invalid action: %s ***' % action
                    print(failure)
                    sys.exit(ECODE.BAD)

        if node.has_select_statement:
            query_obj = DLQuery(result.records)
            tested_records = query_obj.find(select=node.select_statement)
        else:
            tested_records = result.records

        if is_showed:
            Printer.print('User Test Data:')
            print(result.test_data)
            print()

            if result.template:
                Printer.print('Template:')
                print('%s\n' % result.template)
                print()

        lst = ['Parsed Results:']
        if node.has_select_statement:
            lst.append('    SELECT-STATEMENT: %s' % node.select_statement)
        Printer.print(lst)
        if is_tabular:
            Tabular(tested_records).print()
        else:
            pprint(tested_records)
        print()

        if node.is_need_verification and Misc.is_list(tested_records):
            total = len(tested_records)
            op = node.condition
            expected_number = node.expected_condition
            chk = getattr(operator, op)(total, expected_number)
            prefix = '' if chk else '*** CANT BE ***'
            tbl = dict(eq='==', ne='!=', lt='<', le='<=', gt='>', ge='>=')

            lst = ['Verification:',
                   '    CONDITION: %s' % node.condition_data,
                   '    STATUS   : %s' % ('Passed' if chk else 'Failed')]
            Printer.print(lst)

            fmt = '%s (total found records: %s) %s (expected total count: %s)'
            msg = (fmt % (prefix, total, tbl[op], expected_number)).strip()
            print(msg)
            print()

        if node.error:
            lst = ['Error:', '------', '\n'.join(wrap(node.error, width=76)), '']
            Printer.print(lst)
            sys.exit(ECODE.BAD)
        else:
            sys.exit(ECODE.SUCCESS)
